% Implementation of the Lp-RLS algoirthm presented in 
% J. K. Pant, W.-S. Lu, and A. Antoniou, "New improved algorithm for
% compressive sensing based on Lp norm", IEEE Trans. Circuits and
% Systems - II, vol. 61, no. 3. pp. 198-202, Mar. 2014.
%
% This algorithm is suitable for the reconstruction of sparse signals from
% NOISY measurements in compressive sensing. 
%
% Inputs:
%   Phi - a measurement matrix of size (M X N) where M is the number of 
%       measurements and N is signal length.
%   y - measurement vector of length M.
%   x0 - initial estimation of the sparse vector; a zero vector can be 
%       used if estimation is not available.
%   p - a constant real number which can be greater than 0 and
%       smaller than or equal to 1.
%   lam - regularization parameter.
%   steps - number of sequences in the involved sequential optimization.
%
% Output: 
%   xsol - reconstructed signal
%
% Usage:
%   N = 512; M = 200; K = 85;
%   Phi = randn(50,100);
%   Phi = orth(Phi')'; % Constructs measurement matrix with orthonormal rows
%   x = zeros(N,1); 
%   kind = randperm(N);
%   kind = kind(1:K);
%   x(kind) = randn(K,1); % Constructs K-sparse vector x of length N
%   w = randn(M,1);
%   w = w - mean(w);
%   w = (0.01/sqrt(var(w)))*w; % Constructs a noise vector w of length M
%   y = Phi*x+w; % Takes noisy measurement
%
%   x0 = zeros(100,1); 
%   p = 0.1;
%   lam = 1e-5;
%   steps = 80;
%   xsol = solve_LpRLS_problem(randn(50,100),y,x0,p,lam,steps) 
%   % Applies the Lp-RLS algorithm. 
%
%   Written by : Jeevan K. Pant, April 2012.

function [xsol] = solve_LpRLS_problem(Phi,y,x0,e,p,lam,steps)
e1 = e.I; eT = e.T; 
eset = e1*exp(-log(e1/eT)*((1:steps)-1)/(steps-1));
p1 = p.I; pT = p.T; 
pset = p1*exp(-log(p1/pT)*((1:steps)-1)/(steps-1));

xk = x0;
for(i=1:steps)
    e = eset(i); p = pset(i); epsi = 1e-25; 
    x1 = xk.^2+e^2;
    gr1 = Phi'*(Phi*xk-y); 
    gr2 = p*((x1).^(p/2-1)).*xk;
    gk = gr1 + lam*gr2; 
    dk = -gk;
    h2 = p*((x1).^(p/2-2)).*((p-1)*xk.^2 + e^2);
    indh2 = find(h2<1e-5); 
    h2(indh2) = 1e-5;
    denk = norm(Phi*dk)^2 +lam*norm(sqrt(h2).*dk)^2;
    g2 = gk'*gk; 
    ak = g2/denk;
    adk = ak*dk;  
    err = norm(adk);
    k=1;
    while(err>epsi)
        xk = xk + adk; 
        x1 = xk.^2 + e^2;
        gr1 = Phi'*(Phi*xk-y); 
        gr2 = p*((x1).^(p/2-1)).*xk;
        gk = gr1 + lam*gr2;
        h2 = p*((x1).^(p/2-2)).*((p-1)*xk.^2 + e^2);
        indh2 = find(h2<1e-5); 
        h2(indh2) = 1e-5;
        g2_new = gk'*gk; 
        bk = g2_new/g2;
        dk = -gk + bk*dk;
        denk = norm(Phi*dk)^2+lam*norm(sqrt(h2).*dk)^2;
        g2 = g2_new; 
        ak = g2/(denk);
        adk = ak*dk;
        err = norm(adk);
        k = k+1;
        if(k>6+round(i/5)) break; end
    end
    xk = xk+adk;
end
xsol = xk;
